#include <ctime>

// cheatsheet: https://hackingcpp.com/cpp/libs/fmt.html
#include <fmt/core.h>
#include <fmt/chrono.h>
#include <fmt/ostream.h>
#include <fmt/os.h> // for fmt::output_file

// manual: http://tclap.sourceforge.net/manual.html#EXAMPLE
#include <tclap/CmdLine.h>

// manual: https://github.com/SergiusTheBest/plog
#include <plog/Log.h>

/// Main routine
int run(bool val_bool, std::string val_option, float val_pos_arg, std::vector<std::string> val_pos_args) {
	auto out = fmt::output_file("guide.txt");
	// output_file("name", fmt::file::WRONLY | fmt::file::CREATE | fmt::file::APPEND)
	out.print("Don't {}", "Panic");

	fmt::print("val_bool     = {}\n", val_bool);
	fmt::print("val_option   = {}\n", val_option);
	fmt::print("pos_arg      = {}\n", val_pos_arg);
	if (val_pos_args.size() > 0) {
		fmt::print("pos_args     = [\n");
		for (auto arg : val_pos_args) {
			fmt::print("  {}\n", arg);
		}
		fmt::print("]\n");
	} else {
		fmt::print("pos_args     = []\n");
	}


	std::string s = fmt::format("I'd rather be {1} than {0}.", "right", "happy");
	fmt::print("Hello, world!\n");

	{
		using namespace std::literals::chrono_literals;
		fmt::print("Default format: {} {}\n", 42s, 100ms);
		fmt::print("strftime-like format: {:%H:%M:%S}\n", 3h + 15min + 30s);
	}

	return 0;
}

/// Setup routine for the logging interface
int setup(std::string log_level, std::string log_output) {
	auto max_level = plog::debug;
	std::string ulog_level = log_level;
	std::transform(log_level.begin(), log_level.end(), ulog_level.begin(), [](unsigned char c){ return std::toupper(c); });

	// severity levels = {PLOG_VERBOSE, PLOG_DEBUG, PLOG_INFO, PLOG_WARNING, PLOG_ERROR, PLOG_FATAL, PLOG_NONE}
	if (ulog_level == "VERBOSE") {
		max_level = plog::verbose;
	} else if (ulog_level == "DEBUG") {
		max_level = plog::debug;
	} else if (ulog_level == "INFO") {
		max_level = plog::info;
	} else if (ulog_level == "WARNING") {
		max_level = plog::warning;
	} else if (ulog_level == "ERROR") {
		max_level = plog::error;
	} else if (ulog_level == "FATAL") {
		max_level = plog::fatal;
	} else if (ulog_level == "NONE") {
		max_level = plog::none;
	} else {
		fmt::print(stderr, "error: expected maximum severity level, got '{}'\n", log_level);
		fmt::print(stderr, "       admissible values would be {{VERBOSE, DEBUG, INFO, WARNING, ERROR, FATAL, NONE}}\n");
		return 2;
	}

	//plog::init(max_level, "Hello.txt");
	//plog::get()->setMaxSeverity(plog::debug);

	return 0;
}

/// Main routine wrapper with {logging setup, CLI parsing, runtime recording}
int main(int argc, const char* argv[]) {
	// settings
	bool val_bool;
	std::string val_option;
	float val_pos_arg;
	std::string log_level;
	std::string log_output;
	std::vector<std::string> val_pos_args;

	// command line parsing
	try {
		TCLAP::CmdLine cmd("((TODO description))", ' ', "((TODO version))");
		// SwitchArg(FLAG, NAME, DESC, DEFAULT)
		TCLAP::SwitchArg bool_option("r", "reverse", "on or off", false);
		// ValueArg(FLAG, NAME, DESC, REQUIRED, DEFAULT, TYPE_DESC)
		TCLAP::ValueArg<std::string> a_option("a", "option", "some option to take", false, "defval", "string");
		TCLAP::ValueArg<std::string> arg_log_level("l", "log-level", "min logging level like {VERBOSE, DEBUG, WARNING} or {FATAL}", false, "WARNING", "string");
		TCLAP::ValueArg<std::string> arg_log_output("o", "log-output", "logging output; '-' for stdout or provide a filename", false, "-", "string");
		// positional UnlabeledValueArg(NAME, DESC, REQUIRED, DEFAULT, TYPE_DESC)
		TCLAP::UnlabeledValueArg<float> pos_arg("arg", "some positional argument", true, 3.14, "float");
		// positional multi UnlabeledMultiArg(NAME, DESC, REQUIRED, TYPE_DESC)
		TCLAP::UnlabeledMultiArg<std::string> pos_args("files", "files to handle", false, "string");

		cmd.add(bool_option);
		cmd.add(a_option);
		cmd.add(arg_log_level);
		cmd.add(arg_log_output);
		cmd.add(pos_arg);
		cmd.add(pos_args);
		cmd.parse(argc, argv);

		val_bool = bool_option.getValue();
		val_option = a_option.getValue();
		log_level = arg_log_level.getValue();
		log_output = arg_log_output.getValue();
		val_pos_arg = pos_arg.getValue();
		val_pos_args = pos_args.getValue();

	} catch (TCLAP::ArgException &e) {
		fmt::print(stderr, "error: {} for arg {}\n", e.error(), e.argId());
		return 1;
	}

	// application setup incl. logging
	int ret = setup(log_level, log_output);
	if (ret != 0) {
		return ret;
	}

	// print current time before
	/*time_t now;
    char timestamp_buffer[sizeof "2011-10-08T07:07:09Z"];
	time(&now);
    strftime(timestamp_buffer, sizeof timestamp_buffer, "%FT%TZ", gmtime(&now));
	std::string ts1(timestamp_buffer);
	PLOG_DEBUG << "started at " << ts1;*/

	// start actual program logic
	ret = run(val_bool, val_option, val_pos_arg, val_pos_args);

	/*time(&now);
    strftime(timestamp_buffer, sizeof timestamp_buffer, "%FT%TZ", gmtime(&now));
	std::string ts2(timestamp_buffer);
	PLOG_DEBUG << "finished at " << ts2;*/

	return ret;
}
